/* Module      : JaegerDraw.cpp
 * Author      : Robert W. Lindeman
 * Email       : gogo@wpi.edu
 * Course      : IMGD 3000
 *
 * Description : This file contains the routines for drawing the scenes
 *               for the player.
 *
 * Date        : 2006/02/05
 *
 * History:
 * Revision      Date          Changed By
 * --------      ----------    ----------
 * 00.01         2006/02/05    gogo
 * First release.
 *
 * (c) Copyright 2006, Worcester Polytechnic Institute.
 */

/* -- INCLUDE FILES ------------------------------------------------------ */
#include <stdio.h>
#include <stdlib.h>

#ifdef MACOSX
  #include <gl.h>
  #include <glu.h>
  #include <glut.h>
#else
  #include <GL/gl.h>
  #include <GL/glu.h>
  #include <GL/glut.h>
#endif

#include <JaegerDraw.h>
#include <JaegerGlobals.h>
#include <JaegerConstants.h>

using namespace std;

/* -- DATA STRUCTURES ---------------------------------------------------- */

/* -- GLOBAL VARIABLES --------------------------------------------------- */

/* -- LOCAL VARIABLES ---------------------------------------------------- */
static void JaegerDrawMaze( void );
static void JaegerDrawCell( Cell *cell );

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawAxes( void )
 *
 * Description : This function draws orthogonal axes in red, green, and
 *               blue at the current location.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerDrawAxes( void )  {
  GLUquadricObj *qobj;
  GLfloat cylRad = 0.025;
  GLfloat coneRad = 0.075;

  glPushMatrix( );
    // Do the X axis.
    glColor3f( 1.0, 0.0, 0.0 );
    glutSolidSphere( cylRad, 16, 16 );

    glRotated( 90.0, 0.0, 1.0, 0.0 );
    qobj = gluNewQuadric( );
    gluQuadricDrawStyle( qobj, GLU_FILL );
    gluCylinder( qobj, cylRad, cylRad, 1.0, 16, 1 );
    glTranslated( 0.0, 0.0, 1.0 );
    glutSolidCone( coneRad, 0.2, 16, 1 );
  glPopMatrix( );

  glPushMatrix( );
    // Do the Y axis.
    glColor3f( 0.0, 1.0, 0.0 );
    glRotated( -90.0, 1.0, 0.0, 0.0 );
    qobj = gluNewQuadric( );
    gluQuadricDrawStyle( qobj, GLU_FILL );
    gluCylinder( qobj, cylRad, cylRad, 1.0, 16, 1 );
    glTranslated( 0.0, 0.0, 1.0 );
    glutSolidCone( coneRad, 0.2, 16, 1 );
  glPopMatrix( );

  glPushMatrix( );
    // Do the Z axis.
    glColor3f( 0.0, 0.0, 1.0 );
    qobj = gluNewQuadric( );
    gluQuadricDrawStyle( qobj, GLU_FILL );
    gluCylinder( qobj, cylRad, cylRad, 1.0, 16, 1 );
    glTranslated( 0.0, 0.0, 1.0 );
    glutSolidCone( coneRad, 0.2, 16, 1 );
  glPopMatrix( );
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDraw( void )
 *
 * Description : This function draws the entire world.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerDraw( void )  {
  // Hack, so that the camera is placed in the right place.
  Cam.Slide( 0.0, 0.0, 0.0 );

  GLfloat matAmb[]   = { 0.4, 0.4, 0.4, 1.0 };
  GLfloat matDiff[]  = { 0.6, 0.6, 0.6, 1.0 };
  GLfloat matSpec[]  = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat matShine[] = { 50.0 };
  GLfloat lmAmb[]    = { 0.2, 0.2, 0.2, 1.0 };

  glMaterialfv( GL_FRONT, GL_AMBIENT, matAmb );
  glMaterialfv( GL_FRONT, GL_DIFFUSE, matDiff );
  glMaterialfv( GL_FRONT, GL_SPECULAR, matSpec );
  glMaterialfv( GL_FRONT, GL_SHININESS, matShine );

  GLfloat lInten[] = { 0.7, 0.7, 0.7, 1.0 };
  GLfloat lPos[]   = { 100.0, 1000.0, 100.0, 0.0 };
  glLightfv( GL_LIGHT0, GL_POSITION, lPos );
  glLightfv( GL_LIGHT0, GL_DIFFUSE, lInten );
  glLightModelfv( GL_LIGHT_MODEL_AMBIENT, lmAmb );

  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glColor3f( 1.0, 1.0, 1.0 );

  JaegerDrawMaze( );

  // Display the buffer we just rendered into.
  glutSwapBuffers( );
}
  
/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawMaze( void )
 *
 * Description : This function draws the labyrinth.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

static void JaegerDrawMaze( void )  {
  for( int row = 0; row < GertMaze->height; row++ )  {
    for( int col = 0; col < GertMaze->width; col++ )  {
      glPushMatrix( );
        glTranslated( col * CellSide,
                      0.0,
                      row * CellSide );
        JaegerDrawCell( GertMaze->GetCell( row, col ) );
      glPopMatrix( );
    }
  }
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawCell( Cell *cell )
 *
 * Description : This function draws the cell associated with 'cell'.
 *      ________
 *     /       /|
 *    /       / |
 *   /_______/  |
 *   |   /   |  |
 *   |  /    |  /
 *   | /     | /
 *   |/______|/
 * 
 *  There are 15 different cell types (not including the door type).
 *  However, these can be reduced to 5 distinct cell types, with the
 *  others being simple rotations of one of these five. Therefore,
 *  find out which "class" of cell we need to draw, apply any necessary
 *  rotation, and then draw that class of cell.
 *
 * Parameters  : Cell cell : Cell to draw.
 *
 * Returns     : void
 */

static void JaegerDrawCell( Cell *cell )  {

  /*
  if( cell == CurCell )  {
    // Draw axes.
    glPushMatrix( );
      glScaled( 75.0, 75.0, 75.0 );
      JaegerDrawAxes( );
    glPopMatrix( );
  }
  */

  switch( cell->type )  {
    case CELL_CROSS: // This is its own class.
      // Do the pole at the North-West corner.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.6 );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                      -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the pole at the North-East corner.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.6 );
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                      -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the pole at the South-West corner.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.6 );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the pole at the South-East corner.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.6 );
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the floor.
      glPushMatrix( );
        glColor3f( 1.0, 0.7, 0.7 );
        glTranslated( 0.0, -WallThickness * 0.5, 0.0 );
        glScaled( CellSide, WallThickness, CellSide );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      break;

    case CELL_VERT:
    case CELL_HORIZ:
      // Now apply the correct rotation.
      switch( cell->type )  {
        case CELL_VERT:
          break;
        case CELL_HORIZ:
          glRotated( 90.0, 0.0, 1.0, 0.0 );
          break;
      }
      // Do the West wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       0.0 );
        glScaled( WallThickness, CellHeight, CellSide );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the East wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       0.0 );
        glScaled( WallThickness, CellHeight, CellSide );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the floor.
      glPushMatrix( );
        glColor3f( 1.0, 0.7, 0.7 );
        glTranslated( 0.0, -WallThickness * 0.5, 0.0 );
        glScaled( CellSide, WallThickness, CellSide );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      break;

    case CELL_SOUTH_WEST:
    case CELL_SOUTH_EAST:
    case CELL_NORTH_WEST:
    case CELL_NORTH_EAST:
      // Now apply the correct rotation.
      switch( cell->type )  {
        case CELL_SOUTH_WEST:
          break;
        case CELL_SOUTH_EAST:
          glRotated( 90.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_NORTH_WEST:
          glRotated( -90.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_NORTH_EAST:
          glRotated( 180.0, 0.0, 1.0, 0.0 );
          break;
      }
// Adjust the wall sizes for the overlap at corners.
      // Do the West wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       0.0 );
        glScaled( WallThickness, CellHeight, CellSide );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the South wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated(  0.0,
                       ( CellHeight * 0.5 ),
                       ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( CellSide, CellHeight, WallThickness );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the pole at the North-East corner.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.6 );
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                      -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the floor.
      glPushMatrix( );
        glColor3f( 1.0, 0.7, 0.7 );
        glTranslated( 0.0, -WallThickness * 0.5, 0.0 );
        glScaled( CellSide, WallThickness, CellSide );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      break;

    case CELL_NORTH_TEE:
    case CELL_EAST_TEE:
    case CELL_SOUTH_TEE:
    case CELL_WEST_TEE:
      // Now apply the correct rotation.
      switch( cell->type )  {
        case CELL_NORTH_TEE:
          break;
        case CELL_EAST_TEE:
          glRotated( -90.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_SOUTH_TEE:
          glRotated( 180.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_WEST_TEE:
          glRotated( 90.0, 0.0, 1.0, 0.0 );
          break;
      }
      // Do the North wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated(  0.0,
                       ( CellHeight * 0.5 ),
                      -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( CellSide, CellHeight, WallThickness );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the pole at the South-West corner.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.6 );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the pole at the South-East corner.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.6 );
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the floor.
      glPushMatrix( );
        glColor3f( 1.0, 0.7, 0.7 );
        glTranslated( 0.0, -WallThickness * 0.5, 0.0 );
        glScaled( CellSide, WallThickness, CellSide );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      break;

    case CELL_NORTH_DEADEND:
    case CELL_EAST_DEADEND:
    case CELL_SOUTH_DEADEND:
    case CELL_WEST_DEADEND:
      // Now apply the correct rotation.
      switch( cell->type )  {
        case CELL_NORTH_DEADEND:
          break;
        case CELL_EAST_DEADEND:
          glRotated( -90.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_SOUTH_DEADEND:
          glRotated( 180.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_WEST_DEADEND:
          glRotated( 90.0, 0.0, 1.0, 0.0 );
          break;
      }
// Adjust the wall sizes for the overlap at corners.
      // Do the North wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated(  0.0,
                       ( CellHeight * 0.5 ),
                      -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( CellSide, CellHeight, WallThickness );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the West wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       0.0 );
        glScaled( WallThickness, CellHeight, CellSide );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the East wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       0.0 );
        glScaled( WallThickness, CellHeight, CellSide );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      // Do the floor.
      glPushMatrix( );
        glColor3f( 1.0, 0.7, 0.7 );
        glTranslated( 0.0, -WallThickness * 0.5, 0.0 );
        glScaled( CellSide, WallThickness, CellSide );
        glutSolidCube( 1.0 );
      glPopMatrix( );
      break;

    default:
      break;
  }
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawInit( int argc, char *argv[] )
 *
 * Description : This function initializes the graphics state.
 *
 * Parameters  : int argc    : Number of command-line arguments.
 *               char *argv[]: Array of command-line arguments.
 *
 * Returns     : void
 */

void JaegerDrawInit( int argc, char *argv[] )  {
  glutInit( &argc, argv );
  glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
  glutInitWindowSize( ScreenWidth, ScreenHeight );
  glutInitWindowPosition( 100, 100 );
  glutCreateWindow( WindowTitle );

  glutDisplayFunc( JaegerDraw );

  glShadeModel( GL_SMOOTH );
  glEnable( GL_DEPTH_TEST );
  glEnable( GL_NORMALIZE );

  glEnable( GL_LIGHTING );
  glEnable( GL_LIGHT0 );
  glColorMaterial( GL_FRONT, GL_DIFFUSE );
  glEnable( GL_COLOR_MATERIAL );
  glClearColor( 0.1, 0.1, 0.1, 0.0 );

  glMatrixMode( GL_PROJECTION );
  glLoadIdentity( );
  Cam.SetShape( ViewAngle, AspectRatio, NearClip, FarClip );
  Cam.Set( EyePt, InterestPt, UpVec );
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerTurnLeft90( void )
 *
 * Description : This function turns the users view 90 deg. left.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerTurnLeft90( void )  {
  Cam.Yaw( 90.0 );

  switch( Hedi->ori )  {
    case ORI_NORTH:
      Hedi->ori = ORI_WEST;
      break;
    case ORI_WEST:
      Hedi->ori = ORI_SOUTH;
      break;
    case ORI_SOUTH:
      Hedi->ori = ORI_EAST;
      break;
    case ORI_EAST:
      Hedi->ori = ORI_NORTH;
      break;
    default:
      break;
  }
}
  
/* ----------------------------------------------------------------------- */
/* Function    : void JaegerTurnRight90( void )
 *
 * Description : This function turns the users view 90 deg. right.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerTurnRight90( void )  {
  Cam.Yaw( -90.0 );

  switch( Hedi->ori )  {
    case ORI_NORTH:
      Hedi->ori = ORI_EAST;
      break;
    case ORI_WEST:
      Hedi->ori = ORI_NORTH;
      break;
    case ORI_SOUTH:
      Hedi->ori = ORI_WEST;
      break;
    case ORI_EAST:
      Hedi->ori = ORI_SOUTH;
      break;
    default:
      break;
  }
}
  
/* ----------------------------------------------------------------------- */
/* Function    : void JaegerMoveForwardCell( void )
 *
 * Description : This function moves the user 1 cell forward, if
 *               possible.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerMoveForwardCell( void )  {
  switch( Hedi->ori )  {
    case ORI_NORTH:
      Cam.Slide( 0.0, 0.0, -CellSide );
      Hedi->pos[Yidx]--;
      break;
    case ORI_WEST:
      Cam.Slide( 0.0, 0.0, -CellSide );
      Hedi->pos[Xidx]--;
      break;
    case ORI_SOUTH:
      Cam.Slide( 0.0, 0.0, -CellSide );
      Hedi->pos[Yidx]++;
      break;
    case ORI_EAST:
      Cam.Slide( 0.0, 0.0, -CellSide );
      Hedi->pos[Xidx]++;
      break;
    default:
      break;
  }
}
  
/* ----------------------------------------------------------------------- */
/* Function    : void JaegerMoveBackCell( void )
 *
 * Description : This function moves the user 1 cell back, if
 *               possible.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerMoveBackCell( void )  {
  switch( Hedi->ori )  {
    case ORI_NORTH:
      Cam.Slide( 0.0, 0.0, CellSide );
      Hedi->pos[Yidx]++;
      break;
    case ORI_WEST:
      Cam.Slide( 0.0, 0.0, CellSide );
      Hedi->pos[Xidx]++;
      break;
    case ORI_SOUTH:
      Cam.Slide( 0.0, 0.0, CellSide );
      Hedi->pos[Yidx]--;
      break;
    case ORI_EAST:
      Cam.Slide( 0.0, 0.0, CellSide );
      Hedi->pos[Xidx]--;
      break;
    default:
      break;
  }
}
  
/* ----------------------------------------------------------------------- */
/* Function    : void JaegerMainLoop( void )
 *
 * Description : This function enters the event loop for glut.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

int JaegerMainLoop( void )  {
  glutMainLoop( );
  return( 0 );
}

/* ----------------------------------------------------------------------- */
